import pyglet
import gym
from numpy import random, clip
from copy import deepcopy

from utilities import BoundingBox as bb_utils
from utilities import Time, TerminalReward
from worlds import EpisodicEnv
from worlds.events import Event, LTLEvent, ltl_events

from verifier import LTLProperties

# import the env-related methods from env.utils
from ..env_utils import *

# import the classes and a method (config_Pyglet) for graphical output
from ..graphics.pyglet_utils import Text, Image, config_Pyglet

from . import Vehicle, Features
from . import RoadNetworkCross, VehicleNetworkCross

from .constants import *
from ..constants import LEFT


class ScenarioParams:
    """ offers a structured class of parameters for scenario generation.
    The default values and names are described in ScenarioParams.property_defaults.

    """
    property_defaults = {
        #: A tuple (n1, n2), which gives the range of the vehicles on the road,
        # excluding the ego, by n1 <= N <= n2.
        'n_others_range': (0, MAX_NUM_OTHER_VEHICLES),

        #: A tuple or list (x1, x2); the ego's x-position will be sampled uniformly in [x1, x2]
        'ego_pos_range': (rd.h.start_pos, rd.h.stop_region.boundary[0]),

        #: The ego's velocity limit. the ego's speed will be sampled uniformly in [0, ego_v_upper_lim]
        'ego_v_upper_lim': rd.speed_limit,

        #: The index of the lane on which ego will be initialized.
        'ego_lane_i': None,

        #: The range of the other vehicles that are on the horizontal route and are not corresponding to the vehicle
        # generated by the code for a special scenario below.
        'others_h_pos_range': rd.h.pos_range,

        #: The range of the other vehs that are on the vertical route and initially not stopped in the stop region.
        'others_v_pos_range': rd.v.pos_range,

        #: The number of the other vehs that stopped in the stop region, on the other lanes than that of the ego.
        # The maximum is 3 (as there are 4-lanes including ego's). When this is true, the waited_count of the
        # stopped cars (except the ego) in the stop region will be set randomly. This option is useful when
        # generating an initial scenario for WaitThenGo maneuver.
        'n_others_stopped_in_stop_region': 0,

        #: If True, the generated scenario has a vehicle ahead of the ego within self.veh_ahead_max.
        # This option is useful in generating an initial scenario for e.g., Follow.
        'veh_ahead_scenario': False,

        #: If False, there will be no vehicle in the same lane to the ego's.
        'allow_vehicles_in_ego_lane': True,

        #: a constant between 0 and 1, used in resetting the speeds of the vehicles if necessary to make sure that
        # at least an input to the ego exists to avoid collisions or stop-sign violations. The less
        # v_max_multiplier, more conservative scenario is generated (but when v_max_multipler = 1, it could generate
        # a scenario where a collision is inevitable). 0.9 is a value set by the developer through trial-and-error
        # and exhaustive testing.
        'v_max_multiplier': 0.9,

        #: A tuple (da, dtheta), where da is the absolute maximum of the ego's initial perturbations \
        # along the y-axis from the lane centre; dtheta is the absolute maximum of the ego's heading angle
        # perturbation from the horizontal line (the angle parallel to the horizontal lane). The perturbation along
        # the y-axis is sampled from
        #
        #        "uniform(-ego_perturb_lim[0], ego_perturb_lim[0])"
        #
        # within the lane boundary. Similarly, the perturbation of the heading angle is sampled from
        #
        #        "uniform(-ego_perturb_lim[1], ego_perturb_lim[1])."
        #
        # da is restricted by the width of the road, and dtheta the non-holonomic constraint;
        'ego_perturb_lim': (0, 0),

        #: If True, then the ego's heading angle is always initialized towards the lane centre.
        'ego_heading_towards_lane_centre': True,

        #: If True, randomizes the two scenarios: vehicle_ahead_scenario and n_others_stopped_in_stop_region
        'randomize_special_scenarios': False
    }

    def __init__(self, **kwargs):
        for (prop, default) in self.property_defaults.items():
            setattr(self, prop, kwargs.pop(prop, default))

        if len(kwargs) > 0:
            import warnings
            warnings.warn('Unknown keyword argument(s) in ScenarioParams.__init__.')

    def n_others(self):
        """ returns a number randomly in between self.n_others_range[0] and self.n_others_range[0].
        """
        return random.randint(self.n_others_range[0], self.n_others_range[1] + 1)

    def check_and_shuffle(self, n_others):
        """ checks whether all the parameters ara valid for generating a scenario with ego and n_others vehicles;
        if self.randomize_special_scenarios is True, randomly change some parameters; if some parameters are wrong, then
        raises an assertion.

        Args:
            n_others: the number of the other vehicles
        """

        assert 0 <= self.n_others_range[0] <= n_others <= self.n_others_range[1], 'the range of the number of other ' \
                                                                                  'vehicles is wrong.'

        # when self.randomize_special_scenarios = True, shuffle the available scenarios:
        #
        #   - vehicle ahead scenario
        #
        #   - scenario with some other cars stopped in stop region
        #
        # and randomly choose one or a mixture.
        if self.randomize_special_scenarios:
            if n_others >= 1:
                # Activate the comment below
                # if you want to generate more veh_ahead_scenario
                # self.veh_ahead_scenario = random.rand() <= 0.7
                self.veh_ahead_scenario = bool(random.randint(0, 2))
                self.n_others_stopped_in_stop_region = random.randint(0, clip(n_others - self.veh_ahead_scenario, 0, 3) + 1)

            else:
                self.veh_ahead_scenario = False
                self.n_others_stopped_in_stop_region = 0


        assert self.n_others_stopped_in_stop_region <= min(n_others, 3), \
            'n_others_stopped_in_stop_region has to be less than min(3, n_others_range[0]).'

        assert self.veh_ahead_scenario <= n_others, 'the minimum number of the other vehicles (n_others_range[0]) ' \
                                                'has to be at least larger than or equal to 1 for veh_ahead_scenario.'

        assert isinstance(self.veh_ahead_scenario, bool), 'veh_ahead_scenario has to be a boolean.'
        assert isinstance(self.allow_vehicles_in_ego_lane, bool), 'allow_vehicles_in_ego_lane has to be a boolean.'

        assert 0 <= self.v_max_multiplier <= 1, 'v_max_multiplier has to be within the range [0, 1].'
        assert self.ego_v_upper_lim >= 0, 'ego_v_upper_lim has to be non-negative.'
        assert self.ego_lane_i in {0, 1, None}, 'ego_lane_i has to be 0, 1, None.'

        assert all(pl >= 0 for pl in self.ego_perturb_lim), 'the perturbation term in ego_perturb_lim must be non-negative.'

        assert self.ego_perturb_lim[0] <= rd.hlane_width/2, 'the maximum y-perturbation (i.e. ego_perturb_lim[0]) ' \
                                                            'out-of-range.'

        assert self.ego_perturb_lim[1] < np.pi/2, 'the maximum theta-perturbation (i.e., ego_perturb_lim[1]) is ' \
                                                  'out-of-range.'

        assert isinstance(self.ego_heading_towards_lane_centre, bool), 'ego_heading_not_outwards has to be a boolean.'


class SimpleIntersectionEnv(EpisodicEnv):
    """ A Road environment corresponding to an intersection scenario,
    where each route has two lanes with the same direction (left -> right, top -> down).

    This scenario has just one horizontal route and another vertical route.
    Relevant constants are used from the constants file.
    """

    #: The vehicle-to-vehicle desired distance in the aggressive driving policy
    v2v_ref = 6.0

    #: The maximum distance within which self.vehs[i].APs['veh_ahead'] = True
    veh_ahead_max = 40.0

    #: The minimum distance that every two adjacent vehicles has to be
    # apart from each other.
    veh_ahead_min = VEHICLE_SIZE[0]

    #: The maximum of the ego's heading angle
    max_ego_theta = np.pi / 3

    #: The weight vector to calculate the cost.
    # = (y_err, theta_err, theta_dot_err, steer_err, v_err, max(0, v_err), zerk, 1/ttc, u[0], u[1])
    cost_weights = 1.0e-3 * np.array((12.0, 10.0, 2.5, 2.5, 0.0, 30.0, 5.0, 5.0, 1.0, 1.0))

    #: the flag being False when cost_weights is used in the cost method without modification;
    # If True, then the action parts of cost_weights are increased for some edge cases (see the cost method).
    cost_extra_action_weights_flag = False

    #: the extra weights on the actions added to cost_weights in the cost method
    # for some stopped cases when extra_action_weights_flag = True.
    cost_extra_action_weights = (100.0, 40.0)

    # pyglet window object
    window = None

    __action_space = gym.spaces.Box(low=np.array([MIN_ACCELERATION, -MAX_STEERING_ANGLE_RATE]),
                                    high=np.array([MAX_ACCELERATION, MAX_STEERING_ANGLE_RATE]))

    @property
    def action_space(self): return self.__action_space

    # the observation_space is available only after generate_scenario has been called.
    @property
    def observation_space(self):
        feature_len = len(self.get_features_tuple())
        return gym.spaces.Box(low=-np.finfo(np.float32).max, high=np.finfo(np.float32).max, shape=(feature_len,))

    @property
    def ego(self): return self.vehs[EGO_INDEX] if len(self.vehs) > 0 else None

    def ego_ap2int(self):
        return int(self.ego.APs) if self.ego is not None else 0

    def __init__(self, **kwargs):
        """SimpleIntersectionEnv constructor.

        Args:
            options: high level options instance
            n_other_vehs: number of other vehicles
        """

        #: info for displaying on top of ego
        self.ego_info_text = ""

        #: the number of the other vehicles, initialized in generate_scenario
        self.n_other_vehs = None

        #: the list of the vehicles, initialized in generate_scenario
        self.vehs = list()

        #: the vehicle_nework for render, initialized in generate_scenario
        self.vehicle_network = None

        #: the index of the vehicle immediately ahead from the ego,
        #  initialized in generate_scenario when vehicle_ahead_scenario = True
        self.veh_ahead_index = None

        #: the list of indices of the vehicles stopped in the stop region,
        # initialized in generate_scenario when:
        #           n_others_stopped_in_stop_region >= 1
        self.stopped_veh_in_stop_region_indices = list()

        #: the cost evaluated at the previous call of cost(u_ego) when u_ego is not None.
        # initially None, reset to None, but a real number is assigned after every call of cost(u_ego).
        # One can simply call self.cost() without the parameter to get the value of self.__cost.
        self.__cost = None

        #: the reference velocity that the ego tries to follow without exceeding it
        self.v_ref = rd.speed_limit

        #: the target_lane (0: left, 1: right) of the ego to follow
        # (it will be set to the ego's lane by default)
        self.target_lane = LEFT

        # Save the scenario params to reuse it in reset
        self.scenario_params = ScenarioParams(**kwargs)

        super().__init__(time=Time(DT, np.inf))

        events = \
            {
                'goal achieved': Event(lambda: self.ego.x >= rd.h.end_pos and self.ego.v <= 1.1 * rd.speed_limit,
                                       lambda: 100),

                'ego collision': Event(lambda: self.check_ego_collision() or self.is_ego_off_road(), lambda: -100),

                'other veh collision': Event(self.check_other_veh_collisions, lambda: -100),

                'ego off lane': Event(self.is_ego_off_lane, lambda: -100, enable=False),

                'heading angle deviation': Event(lambda: abs(self.ego.theta) > self.max_ego_theta, lambda: -100)
            }

        ltls = LTLProperties(APS_KEYS)
        ltls.add('LTL: intersection rule',  'G( in_intersection => intersection_is_clear)')
        ltls.add('LTL: priority rule',      'G( not in_intersection U highest_priority)')
        ltls.add('LTL: stop region rule',   'G( in_stop_region => (in_stop_region U has_stopped_in_stop_region))')

        events.update(ltl_events(ltls, state=self.ego_ap2int, rewards=lambda: -100))

        self.termination_events.update(events)

        self.generate_scenario()

    def generate_scenario(self, **kwargs):
        """Randomly generate a road scenario with.

            "the N-number of vehicles + an ego vehicle"

        (n_others_range[0] <= N <= n_others_range[1]), all of whose speeds are also set randomly (or deterministically
        depending on the options) between 0 and the speed limit of the road env. The position and speed of the vehicles
        are set (some of them randomly) but not in a way that causes inevitable collision or violation. The ego vehicle
        is always posed on a horizontal lane and there is initially no or only one vehicle posed in the intersection.
        The keyword arguments (**kwargs) contains the values of the attirutes defined in the ScenarioParams class.

        TODO: restructure the code.

        Returns:
            False if the initial state is Terminal, True if not.
        """

        sp = ScenarioParams(**kwargs) if len(kwargs) > 0 else self.scenario_params

        self.n_other_vehs = sp.n_others()
        sp.check_and_shuffle(self.n_other_vehs)

        # a list of vehicles including ego vehicle i = 0 and the others
        # (1 < i <= n_other_vehs <= MAX_NUM_VEHICLES).
        vehs = self.vehs

        # The ego vehicle i = 0 has to be on the horizontal lane.
        vehs.clear()  # Remove the existing vehicles if any.
        vehs.append(Vehicle('h', lane_i=sp.ego_lane_i, v_upper_lim=sp.ego_v_upper_lim))

        ego = self.ego
        ego_lane_i = ego.APs['lane']
        ego_lane_centre = rd.h.lanes[ego_lane_i].centre

        ego_pos_min, ego_pos_max = min(sp.ego_pos_range), max(sp.ego_pos_range)
        others_h_pos_min, others_h_pos_max = min(sp.others_h_pos_range), max(sp.others_h_pos_range)
        others_v_pos_min, others_v_pos_max = min(sp.others_v_pos_range), max(sp.others_v_pos_range)

        ego.x = random.uniform(ego_pos_min, ego_pos_max)

        # sample the ego's y-pos within the lane boundary
        # (ego's center will be in the current lane by the inspection above).
        while True:
            ego.y = ego_lane_centre + random.uniform(-sp.ego_perturb_lim[0], sp.ego_perturb_lim[0])
            sign = np.sign(ego.y - rd.h.lanes[ego_lane_i].centre)

            if sign >= 0 and sp.ego_heading_towards_lane_centre:
                ego.theta = random.uniform(-sp.ego_perturb_lim[1], 0)
            elif sign < 0 and sp.ego_heading_towards_lane_centre:
                ego.theta = random.uniform(0, sp.ego_perturb_lim[1])
            else:  # if not ego_heading_towards_lane_centre:
                ego.theta = random.uniform(-sp.ego_perturb_lim[1], sp.ego_perturb_lim[1])

            if not self.is_ego_off_road():
                break

        ego.init_local_discrete_var()

        # reset v_ref and target_lane to their default values
        self.v_ref = rd.speed_limit
        self.target_lane = ego.APs['lane']

        self.init_APs(False)

        # s is the distance that the vehicle (in this case, the ego)
        # has to inevitably drive for complete stop.
        s = calculate_s(ego.v)
        s = check_init_veh(ego, s, sp.v_max_multiplier)
        s_seq = [s]

        curr_i = 1
        if sp.veh_ahead_scenario:
            while True:
                veh_ahead = Vehicle('h', ego.APs['lane'], rd.speed_limit)
                v2v_dist = random.uniform(self.veh_ahead_min, self.veh_ahead_max)
                veh_ahead.x = ego.x + VEHICLE_SIZE[0] + v2v_dist
                veh_ahead.init_local_discrete_var()

                s = calculate_s(veh_ahead.v)
                s = check_init_veh(veh_ahead, s, sp.v_max_multiplier)

                # Constrain veh ahead's speed in such a way that the ego can't miss it when properly accelerated.
                veh_ahead.v = min(veh_ahead.v, ego.v + (self.veh_ahead_max - v2v_dist) / 5)

                vehs.append(veh_ahead)
                self.init_APs()
                s_seq.append(s)

                if veh_ahead.APs['in_intersection']:
                    if veh_ahead.APs['intersection_is_clear']:
                        self.init_APs(False)
                        break
                    else:
                        vehs.pop()
                        self.init_APs()
                        s_seq.pop()
                        continue
                else:
                    break

            while v2v_dist + s_seq[curr_i] < s_seq[curr_i - 1] + self.veh_ahead_min:
                s_seq[curr_i - 1] = regulate_veh_speed(
                    ego, v2v_dist + s_seq[curr_i] - self.veh_ahead_min, sp.v_max_multiplier)

            curr_i += 1

        # randomly pose the (n_others_stopped_in_stop_region)-number of vehicles
        # with zero speeds in stop regions but not on the same lane as Ego's.
        if sp.n_others_stopped_in_stop_region:
            index_flag = [0] * sp.n_others_stopped_in_stop_region

            # 0s <= max_waited_count * DT <= 1s;
            # this is for the random choice of waited_count of the vehicles
            # which initially have stopped in the stop regions.
            max_waited_count = random.randint(0, 11)
            for i in range(0, sp.n_others_stopped_in_stop_region):

                # Randomly pick an index j = 0, 1, 2. Remember that 3 is the
                # maximum number of vehicles initially forced to be posed
                # in the stop region with zero speeds, i.e.,
                # n_others_stopped_in_stop_region = 0, 1, 2, or 3
                # (see the inspection above).
                while True:
                    j = random.randint(0, sp.n_others_stopped_in_stop_region)
                    # if found an empty intersection, break the loop.
                    if not index_flag[j]:
                        break

                # Now, the stop region corresponding to index_flag[j]-th lane
                # will be occupied.
                index_flag[j] = True

                if j == 0:  # 'h'orizontal lane
                    vehs.append(Vehicle('h', not ego_lane_i, 0))
                    bdy = rd.h.lanes[not ego_lane_i].stop_region.boundary
                    vehs[curr_i + i].x = random.uniform(bdy[0], bdy[1])
                elif j == 1:  # 'v'ertical left lane
                    vehs.append(Vehicle('v', 0, 0))
                    bdy = rd.v.lanes[0].stop_region.boundary
                    vehs[curr_i + i].y = random.uniform(bdy[0], bdy[1])
                elif j == 2:  # 'v'ertical right lane
                    vehs.append(Vehicle('v', 1, 0))
                    bdy = rd.v.lanes[1].stop_region.boundary
                    vehs[curr_i + i].y = random.uniform(bdy[0], bdy[1])
                else:
                    raise RuntimeError("Out of the index.")

                # update the maximum waited_count wrt the vehicles
                # generated at the previous step.
                if i > 0:
                    max_waited_count = max(max_waited_count, self.vehs[curr_i + i - 1].waited_count)

                vehs[curr_i + i].init_local_discrete_var()
                # randomly put waited_count from 0 to 21 + max_waited_count
                vehs[curr_i + i].waited_count = random.randint(0, max_waited_count + 21)
                s_seq.append(0)  # s is zero as the vehicle has zero velocity.

            # Now, it successfully put the vehicles on the stop regions.
            curr_i += sp.n_others_stopped_in_stop_region
            self.init_APs(False)

        for i in range(curr_i, self.n_other_vehs + 1):
            success_flag = False
            while not success_flag:
                vehs.append(Vehicle('h' if random.randint(0, 2) is 0 else 'v', random.randint(0, 2), rd.speed_limit))

                if vehs[i].which_route == 'h':
                    vehs[i].x = random.uniform(others_h_pos_min, others_h_pos_max)

                elif vehs[i].which_route == 'v':
                    vehs[i].y = random.uniform(others_v_pos_min, others_v_pos_max)

                else:
                    raise ValueError("The Vehicle class constructor generates a vehicle "
                                     "which is neither 'h'- or 'v'-type.")

                vehs[i].init_local_discrete_var()
                self.init_APs(False)

                if (vehs[i].APs['lane'] == ego_lane_i and sp.allow_vehicles_in_ego_lane is False) or \
                        (vehs[i].APs['in_intersection'] and not vehs[i].APs['intersection_is_clear']):
                    vehs.pop()
                    self.init_APs(False)
                    continue

                if vehs[i].APs['in_intersection']:
                    for veh in vehs:
                        # making every highest priority False as it is possible
                        # that one of them in the stop region is True.
                        veh.APs['highest_priority'] = False
                    vehs[i].APs['highest_priority'] = True

                s = calculate_s(vehs[i].v)
                s = check_init_veh(vehs[i], s, sp.v_max_multiplier)
                s_seq.append(s)
                self.init_APs(False)

                for j in range(0, i + 1):
                    k, v2v_dist = self.get_v2v_distance(j)
                    if k is not None:
                        if v2v_dist < 0 or (v2v_dist + s_seq[k] < s_seq[j] + self.veh_ahead_min):
                            vehs.pop()
                            s_seq.pop()
                            self.init_APs(False)
                            break
                    if j == i:
                        success_flag = True
                        break

        # Finally checks the constraints that every vehicle has to satisfy, and if not satisfied, modify the speeds.
        success_flag = False
        while not success_flag:
            success_flag = True
            for i, veh in enumerate(vehs):
                k, v2v_dist = self.get_v2v_distance(i)
                if k is None:
                    continue
                elif v2v_dist + s_seq[k] < s_seq[i] + self.veh_ahead_min:
                    s_seq[i] = regulate_veh_speed(veh, v2v_dist + s_seq[k] - self.veh_ahead_min, sp.v_max_multiplier)
                    success_flag = False

        self.init_APs(False)

        # Suffle the vehs in self.vehs, except the ego (i = 0).
        other_vehs_indices = random.permutation(self.n_other_vehs) + 1

        self.veh_ahead_index = other_vehs_indices[0] if sp.veh_ahead_scenario else None

        if sp.n_others_stopped_in_stop_region:
            start_index = sp.veh_ahead_scenario
            self.stopped_veh_in_stop_region_indices = \
                other_vehs_indices[start_index:start_index + sp.n_others_stopped_in_stop_region]

        ego.mode = 'normal'
        other_vehs = []
        for i in other_vehs_indices:
            vehs[i].mode = 'fast'
            other_vehs.append(vehs[i])

        vehs[1:] = other_vehs

        self.__cost = None

        # Reset the variables related to verifier, collision-checking and termination.
        return not super().reset()

    def init_APs(self, update_local_APs=True):
        """Initialize the global (and local if update_local_APs is true) atomic
        propositions (APs) of each vehicle using the current state. This method
        is basically called by "generate_scenario" method. Compared with
        update_APs, the way of setting 'highest_priority' is different.

        Args:
            update_local_APs: True if the local APs are also necessary to
                be initialized; False if not, and only the global APs are
                required to be initialzied.
        """

        if update_local_APs is True:
            for veh in self.vehs:
                veh.update_local_APs()

        self.update_APs(False, False)

        index = None
        for i, veh in enumerate(self.vehs):
            if veh.APs['in_intersection']:
                if index is not None:
                    # if there are multiple cars in the intersection,
                    # then choose only one of them to give the highest
                    # priority, which has already done before in the loop.
                    veh.APs['highest_priority'] = False
                else:
                    index = i
                    veh.APs['highest_priority'] = True
            else:
                veh.APs['highest_priority'] = False

        if index is not None:
            return

        waited_count = -1
        for i, veh in enumerate(self.vehs):
            if waited_count < veh.waited_count:
                waited_count = veh.waited_count
                index = i

        if index is not None:
            self.vehs[index].APs['highest_priority'] = True

    def update_APs(self, update_local_APs=True, update_highest_priority=True):
        """Update the global (and local if update_local_APs is true) atomic
        propositions (APs) of each vehicle using the current state. This method
        is basically called by "step" method below after it updates the
        vehicle's continuous state. By a global AP, it means that the state
        information of the other cars has to be used to update the AP (whereas
        a local AP means that the information of the other cars are not
        necessary). For comparison and more details, see also "updateLocalAPs"
        method in vehicles module.

        Args:
            update_local_APs: True if the local APs are also necessary to
                be update; False if not, and only the global APs are required
                to be updated.
        """

        if update_local_APs is True:
            for veh in self.vehs:
                veh.update_local_APs()

        for i, veh in enumerate(self.vehs):
            veh.APs['intersection_is_clear'] = True
            for j in range(0, len(self.vehs)):
                if j is not i:
                    veh.APs['intersection_is_clear'] &= not self.vehs[j].APs['in_intersection']

            if i == 0:
                veh.APs['target_lane'] = self.target_lane
            else:
                veh.APs['target_lane'] = self.vehs[i].APs['lane']

            j, v2v_dist = self.get_v2v_distance(i)

            if j is None:
                veh.APs['veh_ahead'] = False
            else:
                veh.APs['veh_ahead'] = (0 < v2v_dist <= self.veh_ahead_max)

            if not veh.APs['veh_ahead']:
                veh.APs['veh_ahead_slow'] = False
                veh.APs['veh_ahead_too_close'] = False
            else:
                veh_ahead = self.vehs[j]
                veh.APs['veh_ahead_slow'] = (abs(veh_ahead.v) <= rd.speed_limit / 5)

                s_diff = calculate_s(veh.v if i != 0 else abs(veh.v * np.cos(veh.theta))) - calculate_s(veh_ahead.v)
                veh.APs['veh_ahead_too_close'] = (v2v_dist <= self.veh_ahead_min + s_diff)
                veh.APs['veh_ahead_has_passed_stop_region'] = rd.h.stop_region.boundary[1] < veh_ahead.x

        if update_highest_priority:
            waited_count = -1
            index = None
            for i, veh in enumerate(self.vehs):
                if veh.APs['highest_priority']:
                    if veh.APs['in_stop_region'] or veh.APs['in_intersection']:
                        self.__check_highest_priority()
                        return
                    else:
                        # Flag off the AP when the car gets out of the intersection.
                        veh.APs['highest_priority'] = False

                elif waited_count < veh.waited_count:
                    waited_count = veh.waited_count
                    index = i

            if (waited_count != -1) and (index is not None):
                self.vehs[index].APs['highest_priority'] = True

            self.__check_highest_priority()

    def __check_highest_priority(self):
        """Checks whether there exists only one True highest_priority.
        if there are more than one car which have highest priority, then
        this method raises an exception as shown below.

        """
        exists_highest_priority = False
        for veh in self.vehs:
            if veh.APs['highest_priority']:

                if exists_highest_priority:
                    raise ValueError("The AP highest_priority must be true for only one veh for each time.")

                exists_highest_priority = True

    def get_v2v_distance(self, veh_index=0):
        """Calculates the distance between the target and the nearest adjacent
        car ahead in the same lane. The distance is given as a bumper-to-bumper
        one, so being it not positive means that the two cars collide.

        Args:
            veh_index: the index of the vehicle whose v2v distance is
                returned, ranging from 0 to len(self.vehs)
                 (with 0 for the index of the ego).

        Returns:
            a tuple (i, v2vdist), where v2vdist is the v2v distance
            from the indexed vehicle to the closest vehicle, with its index i,
            on the same route; i is None and v2vdist == np.inf
            when there is no vehicle ahead.
        """
        if veh_index not in range(0, len(self.vehs)):
            raise ValueError("veh_index in SimpleIntersectionEnv.get_v2v_distance out of index")

        tar_veh = self.vehs[veh_index]

        v2v_dist = np.inf
        index = None

        for i, veh in enumerate(self.vehs):

            if i == veh_index:
                continue

            elif (veh.which_route == tar_veh.which_route) and (veh.APs['lane'] == tar_veh.APs['lane']):

                sign = 1 if tar_veh.which_dir == 'forward' else -1

                if tar_veh.which_route is 'h':
                    dist = sign * (veh.x - tar_veh.x)
                    if 0.0 <= dist < v2v_dist:
                        v2v_dist = dist
                        index = i

                else: # the case of "tar_veh.which_route is 'v'":
                    dist = sign * (veh.y - tar_veh.y)
                    if 0.0 <= dist < v2v_dist:
                        v2v_dist = dist
                        index = i

        return index, v2v_dist - VEHICLE_SIZE[0]

    def aggressive_driving_policy(self, veh_index):
        """Implementation of the aggressive driving policy. Calculate the
        (veh_index)-th vehicle's low-level action (a, dot_phi) using aggressive
        driving policy. If using aggressive driving policy for all vehicles,
        there is [has to be] no collision.

        Args:
            veh_index: the index of the vehicle to calculate the action
                (a, dot_phi) via aggressive driving policy.

        Returns:
            current low-level action (a, dot_phi) generated by the
            aggressive driving policy for the current state.
        """
        assert 0 <= veh_index <= len(self.vehs) - 1

        tar_veh = self.vehs[veh_index]

        # Note that:
        # 1) calculate_s calculates the dist. required for the vehicle to
        # completely stop under maximum deceleration; calculation is based
        # on the current speed of the vehicle.
        # 2) the order of IF-THEN rules below is important -- if the
        # order has changed, then it would not behave in the same way.

        i, v2v_dist = self.get_v2v_distance(veh_index)
        s = calculate_s(tar_veh.v)

        stop_region, tar_dist_long = \
            (rd.h.stop_region, tar_veh.x) if tar_veh.which_route is 'h' else (rd.v.stop_region, tar_veh.y)

        # If the clear-ahead-dist (=v2v_dist + s_i) is less than
        # "the v2v reference dist" + "the dist. traveling under the
        # maximum brake", then apply emergency stop.
        if i is not None and (v2v_dist + calculate_s(self.vehs[i].v) < s + self.v2v_ref):
            a = MIN_ACCELERATION  # emergency stop

        # If the vehicle-ahead-dist. is less than v2v_ref + some margin
        # (margin includes the size of the vehicles), and if the vehicle
        # is not in the stop region and has enough dist. ahead to stop in
        # the stop region, then apply THE VEHICLE-AHEAD FOLLOWING
        # CONTROLLER, whose objective is to maintain the v2v distance to
        # v2v_ref. Note that this can be activated even if the vehicle's
        # speed is zero (i.e., s = 0).
        elif i is not None and (v2v_dist <= self.v2v_ref + 6) and \
             ((not tar_veh.APs['has_entered_stop_region'] and stop_region.centre - tar_dist_long > 2 * s) or
              (tar_veh.APs['has_entered_stop_region'] and not tar_veh.APs['in_stop_region'])):

            a = 31.6228 * (v2v_dist - self.v2v_ref) - 7.9527 * (tar_veh.v - self.vehs[i].v) + self.vehs[i].a

        # If the ego is in the stop region and has the highest priority,
        # and if there is no vehicle in the intersection, then increase
        # the velocity (up to speed limit).
        elif tar_veh.APs['in_stop_region'] and tar_veh.APs['highest_priority'] and tar_veh.APs['intersection_is_clear']:

            a = -100.0 * (tar_veh.v - rd.speed_limit)

        # If the ego is in the stop region but neither the highest priority
        # is given nor intersection is clear, then decrease the speed to zero.
        elif tar_veh.APs['in_stop_region'] and \
                not (tar_veh.APs['highest_priority'] and tar_veh.APs['intersection_is_clear']):

            a = -2 * 7.9527 * tar_veh.v - 0.5 * np.clip(20 * tar_veh.v, -1, 1)

        # If the ego is not in the stop region yet but close to it, then
        # apply THE STOP CONTROLLER, whose objective is to stop at the
        # stop region center point.
        elif (not tar_veh.APs['has_entered_stop_region']) and \
                (0 <= stop_region.boundary[1] - tar_dist_long <= 3 * s):

            a = -31.6228 * (stop_region.centre - tar_dist_long) - 7.9527 * tar_veh.v

        # Default maneuver: maintain the speed at the speed limit.
        else:

            a = -10.0 * (tar_veh.v - rd.speed_limit)

        return np.clip(a, MIN_ACCELERATION, MAX_ACCELERATION), 0

    def check_ego_collision(self):
        """Checks if ego collides with any other vehicle in the environment.
        Uses bounding boxes to check if a seperating line exists between given
        vehicle and each of the other vehicles.

        Returns:     True if ego collides with another vehicle, False
        otherwise
        """

        current_veh_bb = self.ego.bounding_box
        for i, veh in enumerate(self.vehs):
            if i is EGO_INDEX:
                continue
            other_veh_bb = veh.bounding_box
            if bb_utils.does_bounding_box_intersect(current_veh_bb, other_veh_bb):
                #print("\nEgo collided with vehicle[", i, "]")
                return True

        return False

    def check_other_veh_collisions(self):
        """Checks if other vehicles other than ego collide in the environment.
        Uses v2v distance for vehicles in same lane, otherwise uses bounding
        boxes Checks all vehicles other than ego.

        Returns: True if a vehicle collides with another vehicle,
                 False otherwise
        """

        for index in range(len(self.vehs)):
            if index is EGO_INDEX:
                continue
            current_veh = self.vehs[index]
            for j in range(index + 1, len(self.vehs)):
                veh = self.vehs[j]
                # check collision using v2v for other vehicles only in same lane
                # and travelling in same direction
                if veh.which_route == current_veh.which_route and \
                        veh.APs['lane'] == current_veh.APs['lane']:
                    if current_veh.which_route is 'h':
                        v2v_dist = abs(veh.x - current_veh.x) - VEHICLE_SIZE[0]
                    else:
                        v2v_dist = abs(veh.y - current_veh.y) - VEHICLE_SIZE[0]

                    if v2v_dist < 0:
                        # TODO: Remove this print or log this somehow
                        #print("\nVehicle[", index, "] collided with vehicle[", j,"]")
                        return True

                #  otherwise, check collision using BB only if at intersection
                elif current_veh.APs['in_intersection']:
                    current_veh_bb = current_veh.bounding_box
                    other_veh_bb = veh.bounding_box
                    if bb_utils.does_bounding_box_intersect(current_veh_bb, other_veh_bb):
                        #print("\nVehicle [", index, "] collided with vehicle [", j,"]")
                        return True

        return False

    def is_ego_off_road(self):
        """Checks if ego goes off-road.

        Returns:     True if ego goes off-road, False otherwise
        """

        # check if ego off-road from horizontal road
        h_offroad = not self.ego.is_within_road_boundaries(rd.h)

        # check if ego off-road from vertical road
        v_offroad = not self.ego.is_within_road_boundaries(rd.v)

        return True if (h_offroad and v_offroad) else False

    def is_ego_off_lane(self):
        """Checks if ego is off the target lane, with boundary checking method
            used in collision-checking. Note that "ego['lane'] == target_lane"
            does not give the correct lane-check since ego['lane'] is calculated
            based on the ego's center position.

        Returns:     True if ego goes off-lane, False otherwise
        """
        return not self.ego.is_within_lane_boundaries()

    def cost(self, u=None):
        """Calculate the driving cost of the ego, i.e., the negative reward for
        the ego-driving.

        Args:
            u: the low-level input (a, steer_rate) to the ego vehicle.

        Returns:
            cost: the weighted sum of squares of the components in
                the vector "vec." Note that a cost is a negative reward.
        """

        if u is not None:

            ego = self.ego

            y_err = ego.y - rd.h.lanes[self.target_lane].centre

            # The target theta error is '0', corresponding to the direction of the
            # horizontal lane. That is,
            #
            #   theta_err = self.ego.theta - 'target_theta' = self.ego.theta - 0.
            #
            theta_err = ego.theta
            steer_err = np.tan(ego.steer) / VEHICLE_WHEEL_BASE
            theta_dot_err = ego.v * steer_err
            zerk = (u[0] - ego.a) / DT
            v_err = ego.v - self.v_ref

            veh_ahead_i, veh_ahead_dist = self.get_v2v_distance()
            if veh_ahead_i is not None and veh_ahead_dist <= (self.veh_ahead_min + self.veh_ahead_max) / 2:
                dv = ego.v - self.vehs[veh_ahead_i].v
                ittc = max(dv, 0) / veh_ahead_dist
            else:
                ittc = 0.0

            vec = (y_err, theta_err, theta_dot_err, steer_err, v_err, max(0, v_err), zerk, ittc, u[0], u[1])

            if self.cost_extra_action_weights_flag:
                cost_weights = list(self.cost_weights)
                if ego.v <= 0 and u[0] < 0:
                    cost_weights[8] += self.cost_extra_action_weights[0] * abs(u[0]) / MAX_ACCELERATION

                if abs(ego.steer) >= MAX_STEERING_ANGLE and np.sign(ego.steer) == np.sign(u[1]):
                    cost_weights[9] += self.cost_extra_action_weights[1] * abs(u[1]) / MAX_STEERING_ANGLE_RATE
            else:
                cost_weights = self.cost_weights

            assert len(cost_weights) == len(vec)

            self.__cost = np.inner(cost_weights, np.square(vec))

        return self.__cost

    def get_features_tuple(self):
        """Get/calculate the features wrt. the current state variables.

        Returns features tuple
        """

        return Features(self).get_features_tuple()

    def step(self, action, action_clipping=True, terminal_reward=True):
        """Gym compliant step function.

        Calculate the next continuous and discrete states of every vehicle
        using the given action u_ego for ego-vehicle and aggressive driving
        policy for the other vehicles. The variables in self.vehs will be
        automatically updated. Then, call the options _step and get the reward
        and terminal values.

        Args:
            action: the low level action (a, dot_phi) of the ego vehicle.
            action_clipping: if True, the action is always projected to the action space;
                             otherwise, raise an assertion error when action doesn't belong to self.action_space
            terminal_reward: if False, the terminal reward is not added to the reward output,
                             but the user deals with it externally by calling self.terminal_reward().

        Returns:
            new_state: the new state of the environment
            R: reward obtained
            terminal: whether or not the current policy finished, that is, its termination_condition() is satisfied;
            info: information variables
        """

        # input clipping function.
        def clip_input(u):
            action_space = self.action_space
            return u if action_space.contains(u) else np.clip(u, action_space.low, action_space.high)

        if action_clipping:
            action = clip_input(action)
        else:
            assert self.action_space.contains(action)

        reward = -self.cost(action)  # a cost is defined as a negative reward.

        for i, veh in enumerate(self.vehs):
            veh.step(action if i == EGO_INDEX else self.aggressive_driving_policy(i))

        self.update_APs(False, True)

        terminal = self.step_termination_events()
        if terminal_reward and terminal:
            reward += self.terminal_reward()

        return self.get_features_tuple(), reward, terminal, self.get_info()

    def reset(self):
        """Gym compliant reset function.

        Reinitialize this environment with whatever parameters it was initially
        called with.

        Returns:
            new_state: the environment state after reset
        """

        while not self.generate_scenario():
            pass

        return self.get_features_tuple()

    def draw(self, info=None):
        """Draw the background, road network and vehicle network. Also show
        information as passed on by info.

        Args:
            info: the information string to be shown on the window
        """

        self.background.draw()

        # draw the road network
        self.road_network.draw()

        # draw the vehicle network
        self.vehicle_network.draw(self.ego_info_text)

        if info is not None:
            label = Text(info, x=10, y=NUM_VPIXELS - 20, multiline=True)
            label.draw()

    def render(self, mode='human'):
        """Gym compliant render function."""

        if mode == 'human':
            if self.window is None:
                # load the background texture
                self.background = Image('background.png', anchor=(0, 0), image_type="bg")

                # load the road network
                self.road_network = RoadNetworkCross('road_texture.png')
                self.vehicle_network = VehicleNetworkCross('car_agent.png', self)
                self.window = config_Pyglet(pyglet, NUM_HPIXELS, NUM_VPIXELS)

            self.window.clear()
            self.window.switch_to()
            self.window.dispatch_events()

            # Text information about ego vehicle's states
            info = "Ego Attributes:\n" \
                   + get_APs(self, EGO_INDEX, 'lane', 'veh_ahead', 'stopped_now', 'in_stop_region',  'in_intersection',
                             'highest_priority', 'intersection_is_clear',
                             'has_stopped_in_stop_region', 'before_but_close_to_stop_region') \
                   + get_veh_attributes(self, EGO_INDEX, 'v', 'a', 'steer', 'steer_rate')

            index, dist = self.get_v2v_distance()
            if index is not None:
                info += get_veh_attributes(self, index, 'v')

            self.draw(info)

            self.window.flip()

        #TODO: make this similar to gym standards
        elif mode == 'rgb_array':
            return self.get_features_tuple()

    def close_window(self):
        if self.window is not None:
            self.window.close()

    def set_ego_info_text(self, info):
        self.ego_info_text = info

    def copy(self):
        # Make a copy of self.env
        # Assumption: all environments ship with this method
        backup_background = self.background
        backup_road_network = self.road_network
        backup_vehicle_network = self.vehicle_network
        backup_window = self.window
        self.background = None
        self.road_network = None
        self.vehicle_network = None
        self.window = None

        new_env = deepcopy(self)

        self.background = backup_background
        self.road_network = backup_road_network
        self.vehicle_network = backup_vehicle_network
        self.window = backup_window

        return new_env

    @property
    def n_others_with_higher_priority(self):
        n_others_with_higher_priority = 0
        for i, veh in enumerate(self.vehs):
            if i == EGO_INDEX:
                pass
            if veh.waited_count > self.ego.waited_count:
                n_others_with_higher_priority += 1

        return n_others_with_higher_priority
